functor ContextFun(structure Throttle' : THROTTLE
		   structure Entry' : ENTRY)
  :> CONTEXT 
  where type Entry.entry = Entry'.entry 
    and type Entry.entry' = Entry'.entry'
    and type Entry.updatemsg = Entry'.updatemsg
    and type Entry.Acl.acl = Entry'.Acl.acl
    and type Entry.Acl.Auth.ident = Entry'.Acl.Auth.ident
    and type Entry.Acl.Rights.rights = Entry'.Acl.Rights.rights
    and type 'a Throttle.throttle = 'a Throttle'.throttle 
    and type AcapTime.acaptime = Entry'.AcapTime.acaptime 
    and type Entry.AcapTime.acaptime = Entry'.AcapTime.acaptime =
struct
  structure Throttle = Throttle'
  structure Entry = Entry'
  structure AcapTime = Entry.AcapTime

  structure Auth = Entry.Acl.Auth
  structure SV = LSyncVar
  structure MC = RMulticast

  exception Modified of AcapTime.acaptime

  exception ContextInconsistency
  exception NotNotifyContext
  exception ContextModified

  datatype outMsg = 
    CNEW of string * Entry.entry * int option
  | CCHANGE of string * Entry.entry * (int * int) option
  | CDELETE of string * Entry.entry * int option
  | CTIME of AcapTime.acaptime
  | CDIE

  type context = {owner : Auth.ident,
		  guard : (string * Entry.entry) -> bool,
		  notify : outMsg CML.chan option,
		  sort : (string * Entry.entry) * (string * Entry.entry) -> 
 		             bool, (* will default to sorting entry names *)
		  data : (string * Entry.entry) list, 
		               (* maintained in sorted order *)
		  hold : (string * Entry.entry) list,
		  enumerate : bool,
		  die : unit SV.ivar,
		  listen : Entry.updatemsg MC.port list,
		  lastupdate : AcapTime.acaptime} SV.mvar

  fun defaultSort ((n1 : string, _), (n2 : string, _)) = (n1 > n2)

  (* create the context *)
  (* val create : Entry.Acl.Auth.ident -> (Entry.entry -> bool)
   -> sort option * outMsg CML.chan option * bool -> context *)
  fun create owner guard (sort, notify, enumerate) =
    (SV.mVarInit {owner=owner,
		  guard=guard o #2,
		  notify=notify,
		  sort=getOpt(sort, defaultSort),
		  data=nil,
		  hold=nil,
		  enumerate=enumerate,
		  die=SV.iVar (),
		  listen=nil,
		  lastupdate=AcapTime.old}) : context
     
  (* no notifications & no guard function; 
   doesn't resort until "doneInitial" is called *)
  (* val fastInsert : context -> string * Entry.entry -> unit *)
  fun fastInsert (context : context) new =
    let
      val ({owner,guard,notify,sort,data,hold,
	    enumerate,die,listen,lastupdate}) =
	SV.mTake context
    in
      SV.mPut (context, 
	       {owner=owner,guard=guard,notify=notify,sort=sort,
		enumerate=enumerate,die=die,data=data,hold=new::hold,
		listen=listen,lastupdate=lastupdate})
    end
  
  (* inserts in sorted order, returns position *)
  fun insertInList (sort, list, new) =
    let
      fun il (nil, acc, pos) = (List.rev (new::acc), pos)
	| il (l as a::ll, acc, pos) = 
	if sort(a, new) then (List.revAppend (acc, new::l), pos)
	else il (ll, a::acc, pos+1)
    in
      il (list, [], 1)
    end

  (* deletes from the sorted list, returns position *)
  fun deleteInList (sort, list, old as (name, _)) =
    let
      fun dl (nil, acc, pos) = raise ContextInconsistency
	| dl (l as (a as (n, _))::ll, acc, pos) = 
	      if n = name then (List.revAppend(acc, ll), pos)
	      else dl (ll, a::acc, pos+1)
    in
      dl (list, [], 1)
    end

  (* attempt to insert new entry into context; give notification *)
  (* val insert : context -> string * Entry.entry -> unit *)
  fun insert (context : context) new =
    let
      val (c as {owner,guard,notify,sort,data,hold,
		 enumerate,die,listen,lastupdate}) = SV.mTake context
    in
      if guard new then
	(* add it:
	 - put it (sorted) into data
              - can NOT change an existing entry
	 - get the time
	 - if notify, throw an outMsg appropriately *)
	let
	  val (name, entry) = new

	  val (data, pos) = insertInList (sort, data, new)

	  val lastupdate = Entry.gettime entry
	in
	  (case notify of
	     NONE => ()
	   | (SOME chan) => CML.send (chan, (CNEW(name, entry, 
						  if enumerate then SOME pos
						  else NONE))));
	     SV.mPut (context, 
		      {owner=owner,guard=guard,notify=notify,sort=sort,
		       data=data,hold=hold,enumerate=enumerate,die=die,
		       listen=listen,lastupdate=lastupdate})
	end
      else (* not in our scope; it dies here *)
	SV.mPut (context, c)
    end

  (* change an entry that may or may not be in the context *)
  (* val change : context -> (string * Entry.entry) * (string * Entry.entry) 
    -> unit *)
  fun change (context : context) (old, new) =
    let
      val (c as {owner,guard,notify,sort,data,hold,
		 enumerate,die,listen,lastupdate}) = SV.mTake context
    in
      if guard old then
	if guard new then (* CHANGE *)
	  let
	    val (name, newentry) = new

	    val (data, oldpos) = deleteInList (sort, data, old)
	    val (data, newpos) = insertInList (sort, data, new)

	    val lastupdate = Entry.gettime newentry
	  in
	  (case notify of
	     NONE => ()
	   | (SOME chan) => CML.send (chan, (CCHANGE(name, newentry, 
						  if enumerate then 
						     SOME (oldpos, newpos)
						  else NONE))));
	     SV.mPut (context, 
		      {owner=owner,guard=guard,notify=notify,sort=sort,
		       die=die, enumerate=enumerate, data=data,hold=hold,
		       listen=listen,lastupdate=lastupdate})
	  end
	else (* DELETE *)
	  let
	    val (name, entry) = new
	      
	    val (data, pos) = deleteInList (sort, data, old)
	      
	    val lastupdate = Entry.gettime entry
	  in
	  (case notify of
	     NONE => ()
	   | (SOME chan) => CML.send (chan, (CDELETE(name, entry, 
						  if enumerate then SOME pos
						  else NONE))));
	     SV.mPut (context, 
		      {owner=owner,guard=guard,notify=notify,sort=sort,
		       die=die, enumerate=enumerate, data=data,hold=hold,
		       listen=listen,lastupdate=lastupdate})
	  end
      else 
	if guard new then (* ADD *)
	  let
	    val (name, entry) = new
	      
	    val (data, pos) = insertInList (sort, data, new)
	      
	    val lastupdate = Entry.gettime entry
	  in
	  (case notify of
	     NONE => ()
	   | (SOME chan) => CML.send (chan, (CNEW(name, entry, 
						  if enumerate then SOME pos
						  else NONE))));
	     SV.mPut (context, 
		      {owner=owner,guard=guard,notify=notify,sort=sort,
		       die=die,enumerate=enumerate,data=data,hold=hold,
		       listen=listen,lastupdate=lastupdate})
	  end
	else (* NOOP *)
	  SV.mPut (context, c)
    end

  (* delete an entry that may or may not be in the context *)
  (* val delete : context -> (string * Entry.entry) -> unit *)
  fun delete (context : context) old =
    let
      val (c as {owner,guard,notify,sort,data,hold,
		 enumerate,die,listen,lastupdate}) = SV.mTake context
    in
      if guard old then
	let
	  val (name, entry) = old

	  val (data, pos) = deleteInList (sort, data, old)

	  val lastupdate = Entry.gettime entry
	in
	  (case notify of
	     NONE => ()
	   | (SOME chan) => CML.send (chan, (CDELETE(name, entry, 
						  if enumerate then SOME pos
						  else NONE))));
	     SV.mPut (context, 
		      {owner=owner,guard=guard,notify=notify,sort=sort,
		       enumerate=enumerate,die=die,data=data,hold=hold,
		       listen=listen,lastupdate=lastupdate})
	end
      else (* not in our scope *)
	SV.mPut (context, c)
    end

  (* ask the context to send a CTIME message *)
  (* val updatecontext : context -> unit *)
  fun updatecontext (context : context) =
    let
      val ({owner,guard,notify,sort,data,hold,enumerate,die,listen,lastupdate}) =
	SV.mGet context
    in
      case notify of
	 NONE => raise NotNotifyContext
       | (SOME chan) => CML.send (chan, (CTIME lastupdate))
    end

  fun appTab f l =
    let
      fun appTab' (nil, n) = ()
	| appTab' (a::l, n) = ((f (n,a)) : unit; appTab' (l, n+1))
    in
      appTab' (l, 1)
    end

  (* search the current context *)
  (* val search : context -> ((int option * Entry.entry) -> bool) * 
      (string * Entry.entry) Throttle.throttle * AcapTime.acaptime -> unit *)
  fun search (context : context) (sfun, out, time) =
    let
      val out' = Throttle.call out

      val ({owner,guard,notify,sort,data,hold,
	    enumerate,die,listen,lastupdate}) = SV.mGet context

      val win = if enumerate then 
                  fn (pos, (thing as (_, entry))) => 
		       if sfun (SOME pos, entry) then out' thing else ()
		else
		  fn (_, (thing as (_, entry))) =>
		       if sfun (NONE, entry) then out' thing else ()
    in
      if AcapTime.compare (time, lastupdate) = LESS then
	raise ContextModified
      else
	(appTab win data;
	 Throttle.done out)
    end

  (* add a new dataset to listen to *)
  (* val addDSet : context -> Entry.updatemsg MC.port -> unit *)
  fun addDSet (context : context) port =
    let
      val ({owner,guard,notify,sort,data,hold,enumerate,
	    die,listen,lastupdate}) = SV.mTake context
      val evt = MC.recvEvt port
    in
      SV.mPut (context, 
	       {owner=owner,guard=guard,notify=notify,sort=sort,
		die=die,enumerate=enumerate,data=data,hold=hold,
		listen=port::listen,lastupdate=lastupdate})
    end

  (* done with the context *)
  (* val destroy : context -> unit *)
  fun destroy (context : context) =
    let
      val ({owner,guard,notify,sort,data,hold,enumerate,
	    die,listen,lastupdate}) = SV.mGet context

      val _ = 
	case notify of 
	  NONE => ()
	| (SOME chan) => CML.send (chan, CDIE)

      val _ = List.app MC.release listen
    in
      SV.iPut (die, ())
    end

  fun listener (args as (ctx, evts)) =
    case CML.select (evts) of
      (SOME (Entry.NEW new)) => (insert ctx new; listener args)
    | (SOME (Entry.CHANGE chg)) => (change ctx chg; listener args)
    | (SOME (Entry.DELETE old)) => (delete ctx old; listener args)
    | NONE => ()

  fun startListener (ctx, die, listen) =
    let
      val dieevt = CML.wrap (SyncVar.iGetEvt (die), fn () => NONE)
      val evts = List.map (fn p => CML.wrap (MC.recvEvt p, SOME)) listen
    in
      ignore (CML.spawnc listener (ctx, dieevt::evts))
    end

  (* has side-effect of starting a thread to listen for changes *)
  (* val doneInitial : context -> unit *)
  fun doneInitial (context : context) =
    let
      val ({owner,guard,notify,sort,data,hold,
	    enumerate,die,listen,lastupdate}) =
	SV.mTake context

      val data = ListMergeSort.sort (sort) (data @ hold)

      val _ = startListener (context, die, listen)
    in
      SV.mPut (context, 
	       {owner=owner,guard=guard,notify=notify,sort=sort,
		data=data,hold=nil,enumerate=enumerate,die=die,
		listen=listen,lastupdate=lastupdate})
    end
end
